"""
Ramda mapping functions
"""
from collections import OrderedDict
from typing import Iterable, Union, Any, Mapping, MutableMapping, Dict, Callable

from jcramda.base.comparison import is_a_dict, is_a_func, is_a_int, is_a_mapper, is_simple_iter
from jcramda.base.sequence import nth
from jcramda.core import (curry, delitem, co, first, foreach, setitem, not_a, not_none,
                          of, all_, truth, is_a, when, eq, fold, mapof, filter_, filter_of)

__all__ = (
    'prop',
    'loc',
    'obj',
    'keys',
    'values',
    'remove',
    'des',
    'map_with_keys',
    'map_update',
    'map_apply',
    'firstitem',
    'firstvalue',
    'firstkey',
    'obj_zip',
    'update_path',
    'key_map',
    'trans_keys',
    'popitem',
    'sorted_by_key',
    'assign',
    'mstrip',
    'strip_none',
    'strip_empty',
    'flat_concat',
    'orderby',
    'ordered_key',
    'ordered_value',
    'pickall',
    'pick',
    'invert',
    'key_tree',
    'keys_eq',
    'itempath',
    'path_eq',
    'where',
    'where_eq',
    'pluck',
    'depop',
    'map_dict',
)

not_dict = not_a(dict)


@curry
def prop(prop_name: str, mapper: Mapping, default=None):
    result = mapper
    for key in prop_name.split('.'):
        result = result.get(key, default)
    return result


@curry
def loc(prop_name, mapper):
    if hasattr(mapper, 'loc'):
        return mapper.loc[prop_name]
    if is_a_int(prop_name):
        prop_name = nth(prop_name)(mapper)
    if hasattr(mapper, 'get'):
        return mapper.get(prop_name)
    return None


@curry
def obj(_keys: Union[str, Iterable[Any]], _values):
    if isinstance(_keys, str):
        return {_keys: _values}
    return dict(zip(_keys, _values))


def keys(mapper: Mapping):
    return mapper.keys()


def values(mapper: Mapping):
    return mapper.values()


def items(mapper: Mapping):
    return mapper.items()


@curry
def des(_keys: Iterable, mapper: Mapping):
    return of(loc(k, mapper) for k in _keys)


@curry
def pickall(_keys: Iterable, mapper: Mapping):
    return dict(zip(_keys, des(_keys, mapper)))


@curry
def pick(_keys: Iterable, mapper: Mapping):
    return dict(
        # filter(None, map(lambda k: (k, loc(k, mapper)) if k in mapper else None, _keys))
        ((k, loc(k, mapper)) for k in _keys if k in mapper)
    )


@curry
def map_update(f, d, v):
    """

    :param f: (v) -> dict
    :param d: dict  updated dict
    :param v: a value
    :return: dict
    """
    d.update(f(v))
    return d


map_with_keys = curry(lambda func, _keys, mapper: map(func, des(_keys, mapper)))

# ( f: (x) -> dict, seqs: [x1, x2 ... xn] ) -> { **f(x1), **f(x2) ... **f(xn) }
map_apply = curry(lambda f, seqs: fold(map_update(f), {}, seqs))


@curry
def remove(_keys: Iterable, mapper: MutableMapping):
    foreach(lambda k: delitem(k, mapper), _keys)
    return mapper


# (d: dict) -> d.values()[0]
firstitem = co(first, items)
firstvalue = co(first, values)
firstkey = co(first, keys)


@curry
def obj_zip(_keys: Iterable, _values: Iterable):
    return dict(zip(_keys, _values))


@curry
def update_path(_path: str, upset, d: MutableMapping):
    paths = tuple(reversed(_path.split('.')))
    new_value = upset(prop(_path, d)) if is_a_func(upset) else upset
    query = fold(lambda r, k: {k: r}, {paths[0]: new_value}, paths[1:])
    d.update(query)
    return d


@curry
def key_map(fn, d: Mapping):
    if not is_a_dict(d):
        return d
    r: Dict = {}
    foreach(lambda k: setitem(r, fn(k), d[k]), d.keys())
    return r


@curry
def trans_keys(key_fn, d, deep=False):
    if is_simple_iter(d):
        result = (trans_keys(key_fn, item, deep) if is_a_dict(item) else item for item in d)
    elif is_a_dict(d):
        result = key_map(key_fn, d)
    else:
        return d

    if deep:
        for key in result:
            result[key] = trans_keys(key_fn, result[key], deep)

    return result


@curry
def popitem(key, d: MutableMapping):
    return d.pop(key)


@curry
def sorted_by_key(key_f, d, reverse=False):
    return OrderedDict(sorted(d.items(), key=key_f, reverse=reverse))


def assign(*args):
    mappers = filter_of(is_a_dict, args)
    if not mappers:
        return {}
    return dict(zip(mapof(keys, mappers), mapof(values, mappers)))


@curry
def mstrip(f, mapper):
    return type(mapper)(filter_(f, mapper.items()))


strip_none = mstrip(lambda item: not_none(item[1]))
strip_empty = mstrip(lambda item: truth(item[1]))


_need_checked_type = is_a((Mapping, list, tuple))


def flat_concat(*args, **kwargs):
    """ 平铺传入的字典
    如果传入参数中有Mapping，则只处理Mapping
    如果传入参数中没有Mapping，则会处理 Sequence[Mapping]
    否则什么都不处理
    :param args: Union[list, dict]
    :param kwargs:
    :return:
    """
    dicts = filter_(all_([is_a_mapper, truth]), of(args))
    lists = of(*filter_(all_([is_simple_iter, truth]), args))

    merged = assign(*dicts, kwargs)
    if merged:
        return dict(
            ((k, flat_concat(v) if _need_checked_type(v) else v) for (k, v) in merged.items() if v)
        )

    return of(flat_concat(x) if _need_checked_type(x) else x for x in lists) \
        or {*filter_(truth, args)}


@curry
def orderby(key_f, d: dict, reverse=False):
    """
    对字典进行排序

    > r = orderby(lambda item: item[1], {'a':3, 'b':1})
    > print(r)
    > OrderedDict([('b', 1), ('a', 3)])

    :param key_f: Tuple[_K, _V] -> Comparable
    :param d: dict
    :param reverse: bool
    :return:
    """
    return OrderedDict(sorted(d.items(), key=key_f, reverse=reverse))


ordered_key = orderby(None)
ordered_value = orderby(lambda x: x[1])


def invert(d: Mapping):
    """
    invert a mapper's key and value
    :param d:
    :return:
    """
    r: Dict = {}
    for k, v in d.items():
        r[v] = of(r[v], k) if v in r else k
    return r


def key_tree(d, prefix=''):
    result = []
    for k, v in when(
            (is_a_mapper, items),
            (is_a(list), enumerate),
            else_=[])(d):
        key_node = f'{prefix}{k}'
        result.append(key_node)
        result += key_tree(v, prefix=f'{key_node}.')
    return result


@curry
def itempath(paths: Union[str, Iterable], mapping):
    r = mapping
    for x in paths.split('.') if isinstance(paths, str) else paths:
        r = when((is_a_mapper, loc(x)), (is_a(Iterable), nth(x)))(r)
        if r is None:
            return None
    return r


@curry
def path_eq(paths: Union[str, Iterable], pred, mapping):
    check = pred if is_a_func(pred) else eq(pred)
    return check(itempath(paths, mapping))


@curry
def pluck(key, mapper: Mapping, *args):
    return tuple(prop(key, d) for d in (mapper, *args))


@curry
def keys_eq(d1, d2) -> bool:
    return keys(d1) == keys(d2)


@curry
def where(pred: Dict[Any, Callable[[Any], bool]], mapping: Mapping) -> bool:
    return all(k in mapping and f(mapping[k]) for (k, f) in pred.items())


@curry
def where_eq(pred, mapping):
    if keys_eq(pred, mapping):
        return where(pred, mapping)
    return False


@curry
def depop(_keys, mapping):
    return tuple(mapping.pop(key) for key in _keys if key in mapping)


@curry
def map_dict(f, mapping):
    return dict(f(k, v) for (k, v) in mapping.items())



from typing import Optional, Any, TypeVar, Generic, Callable
from abc import ABC, abstractmethod
from jcramda.core import is_none, is_a
from jcramda.base import nth, prop


__all__ = (
    'Just',
    'Nothing',
    'Maybe',
)


_TV = TypeVar('_TV')


class NothingValueError(RuntimeError):
    ...


class Maybe(Generic[_TV], ABC):
    def __new__(cls, *args, **kwargs):
        if cls is Nothing:
            return cls()
        v = nth(0, args)
        empty_f = prop('empty_func', kwargs, default=is_none)
        if is_a(Nothing, v) or empty_f(v):
            return super(Maybe, cls).__new__(Nothing)
        if cls is Maybe:
            cls = Just
        return super(Maybe, cls).__new__(cls, *args, **kwargs)
        

    @abstractmethod
    def get(self) -> _TV: ...

    @abstractmethod
    def empty(self) -> bool: ...

    @abstractmethod
    def get_or(self, or_value: _TV): ...

    @abstractmethod
    def get_else(self, else_func: Callable[[], _TV]): ...

    @abstractmethod
    def map(self, func: Callable[[_TV], Any]): ...

    @abstractmethod
    def flatmap(self, func: Callable[[_TV], Any]): ...

    @staticmethod
    def of(v, empty_func=is_none):
        while is_a_factor(v):
            if v.empty:
                return v
            v = v.get()
        return Just(v, empty_func) if not empty_func(v) else Nothing()


class Nothing(Maybe):

    def get(self):
        raise NothingValueError('a nothing value')

    def empty(self) -> bool:
        return True

    def get_or(self, or_value):
        return or_value

    def get_else(self, else_func):
        return else_func()

    def map(self, func):
        return self

    def flatmap(self, func):
        return self
    
    def __repr__(self):
        return 'Maybe::Nothing'


is_a_factor = is_a(Maybe)
is_nothing = is_a(Nothing)


class Just(Maybe):

    def __init__(self, v: Optional[Any], empty_func=is_none):
        self._value = v
        self._empty_func = empty_func

    @property
    def empty(self) -> bool:
        return self._empty_func(self._value)

    def get(self):
        if self.empty:
            raise NothingValueError('a empty value')
        return self._value

    def get_or(self, or_value):
        return or_value if self.empty else self._value

    def get_else(self, else_func):
        return else_func() if self.empty else self._value

    def map(self, func):
        return Just(func(self._value), empty_func=self._empty_func)

    def flatmap(self, func):
        result = func(self._value)
        while is_a_factor(result):
            result = result.get()
        return Just(result, empty_func=self._empty_func)

    def __repr__(self):
        return f'Maybe::Just({self._value})'



import operator as _op
from functools import reduce
from typing import Tuple, Callable, Any, Iterable
from itertools import islice
from ._curry import curry, flip, _


__all__ = (
    'lt', 'le', 'eq', 'ne', 'ge', 'gt', 'not_', 'truth', 'false_', 'is_', 'is_not', 'is_a', 'not_a',
    'not_none', 'is_none', 'between', 'not_in', 'cmp_range', 'clamp', 'dec', 'inc',
    'add', 'sub', 'and_', 'floordiv', 'div', 'inv', 'lshift', 'mod', 'mul', 'matmul',
    'neg', 'or_', 'pos', 'pow_', 'xor', 'concat', 'in_', 'countof', 'delitem', 'getitem',
    'index', 'setitem', 'attr', 'props', 'bind', 'eq_by', 'case', 'indexall',
    'identity', 'when', 'always', 'if_else', 'all_', 'any_', 'all_pass', 'one_pass', 'default_to', 'import_',
    'from_import_as', 'eq_attr', 'eq_prop', 'has_attr', 'try_catch', 'else_to',
)


# Comparison ===========================

eq = curry(_op.eq)
ne = curry(_op.ne)
lt = flip(_op.lt)
le = flip(_op.le)
ge = flip(_op.ge)
gt = flip(_op.gt)


@curry
def eq_by(func, a, b):
    return func(a) == func(b)


@curry
def cmp_range(rng: range, v):
    if v in rng:
        return 0
    return 1 if v >= range.stop else -1


@curry
def between(_min, _max, v):
    """
    _min <= v < _max
    :param _min:
    :param _max:
    :param v:
    :return:
    """
    return cmp_range(range(_min, _max), v) == 0


@curry
def clamp(_min, _max, v):
    result = cmp_range(range(_min, _max), v)
    if result == 0:
        return v
    return _min if result < 0 else _max


# Logical ===============================

not_ = _op.not_
truth = _op.truth
is_ = flip(_op.is_)
is_not = flip(_op.is_not)
is_a = curry(lambda types, obj: isinstance(obj, types))
not_a = curry(lambda types, obj: not isinstance(obj, types))
is_none = is_(None)
not_none = is_not(None)
and_ = curry(lambda a, b: a and b)
or_ = curry(lambda a, b: a or b)



def false_(a):
    return False if a else True

# Math


@curry
def add(a, b):
    return a + b


sub = flip(_op.sub)
dec = sub(1)
inc = add(1)
floordiv = flip(_op.floordiv)
div = flip(_op.truediv)
inv = _op.inv
lshift = flip(_op.lshift)
mod = flip(_op.mod)
mul = curry(_op.mul)
matmul = curry(_op.matmul)
neg = _op.neg
pos = _op.pos
pow_ = flip(_op.pow)
xor = curry(_op.xor)


# Sequence Base
@curry
def concat(a, b, *args):
    return reduce(_op.concat, (a, b, *args))


in_ = curry(_op.contains)
not_in = curry(lambda a, b: b not in a)
countof = flip(_op.countOf)
delitem = flip(_op.delitem)
getitem = flip(_op.getitem)
setitem = curry(lambda d, key, value: _op.setitem(d, key, value))

attr = _op.attrgetter
props = _op.itemgetter
bind = _op.methodcaller


# iadd = flip(_op.iadd)
# iand = flip(_op.iand)
# iconcat = flip(_op.iconcat)
# ifloordiv = flip(_op.ifloordiv)
# ilshift = flip(_op.ilshift)
# imod = flip(_op.imod)
# imul = flip(_op.imul)
# imatmul = flip(_op.imatmul)
# ior = flip(_op.ior)
# ipow = flip(_op.ipow)
# irshift = flip(_op.irshift)
# isub = flip(_op.isub)
# idiv = flip(_op.itruediv)
# ixor = flip(_op.ixor)


# customs ==============================
@curry
def index(x, xs, start=0, end=None):
    try:
        return _op.indexOf(tuple(islice(xs, start, end)), x)
    except ValueError:
        return None


@curry
def indexall(x, xs, start=0, end=None):
    result = []
    for i, v in enumerate(islice(xs, start, end)):
        if v == x:
            result.append(i)
    return iter(result)


@curry
def identity(f, *args, **kw):
    if callable(f):
        return f(*args, **kw)
    return f


def when(*cases: Tuple[Callable, Any], else_=None):
    # noinspection PyBroadException
    def cond(value):
        for f, elem in cases:
            try:
                if f(value):
                    return identity(elem, value)
            except Exception:
                continue
        return identity(else_, value)
    return cond


@curry
def case(cases: dict, v, default=None):
    return identity(cases.get(v), v) or default


@curry
def always(x, _):
    return x


@curry
def if_else(pred, success, failed, value):
    return success(value) if pred(value) else failed(value)


# noinspection PyBroadException
@curry
def try_catch(f, failed, value):
    try:
        return f(value)
    except Exception:
        return failed(value)


@curry
def all_(funcs: Iterable[Callable[[Any], bool]], v):
    return all(f(v) for f in funcs)


@curry
def one_pass(func, iterable) -> bool:
    for x in iterable:
        if func(x):
            return True
    return False


@curry
def all_pass(func, iterable) -> bool:
    for x in iterable:
        if not func(x):
            return False
    return True


@curry
def any_(funcs: Iterable[Callable[[Any], bool]], v):
    for f in funcs:
        if f(v):
            return True
    return False


@curry
def default_to(df, raw):
    return raw or df


@curry
def else_to(else_f, raw, args=(), kwargs={}):
    return raw or else_f(*args, **kwargs)


def import_(module_name, package=None):
    from importlib import import_module
    try:
        return import_module(module_name, package)
    except TypeError:
        return None


@curry
def from_import_as(_name: str, package=None):
    if package:
        name, module_name = (_name, package)
    else:
        name, module_name = _name.rsplit('.', 1)
    super_module = import_(module_name)
    return getattr(super_module, name, None)


@curry
def eq_attr(attr_name, o1, o2):
    attr_getter = _op.attrgetter(attr_name)
    return attr_getter(o1) == attr_getter(o2)


@curry
def eq_prop(prop_name, s1, s2):
    proper = _op.itemgetter(prop_name)
    return proper(s1) == proper(s2)


@curry
def has_attr(attr_name, obj, pred=None):
    if hasattr(obj, attr_name):
        return pred(getattr(obj, attr_name)) if callable(pred) else True
    return False



from abc import ABCMeta
from inspect import signature, Parameter, getfullargspec, FullArgSpec
from functools import wraps, partial
from typing import Callable, List, TypeVar, Union

__all__ = (
    'EmptyParam',
    'curry',
    'flip',
    'is_curried',
    '_',
    'CurriedF',
)
_CRT = TypeVar('_CRT', covariant=True)
CurriedF = Union[Callable[..., _CRT], _CRT]
EmptyParam = Parameter.empty
_ = EmptyParam


def _count_args(spec: FullArgSpec):
    return len(spec.args) + len(spec.kwonlyargs)


def update_args(fn, args, kws):
    old_args = list(args)
    new_kws = kws or {}
    new_args = []
    if isinstance(fn, partial):
        new_args = (old_args.pop(0) if x is _ else x for x in fn.args)
        if fn.keywords:
            new_kws.update(fn.keywords)
        fn = fn.func

    return partial(fn, *new_args, *old_args, **new_kws)


def is_filled(fn: partial, spec):
    return _ not in fn.args and \
        len(spec.args or ()) <= len(fn.args) + len(spec.defaults or ()) and \
        {*spec.kwonlyargs} <= {*fn.keywords, *(spec.kwonlydefaults or ())}


def _need_curried_f(spec):
    return _count_args(spec) > 1 or spec.varargs or spec.varkw


def make_curry(fn, spec):
    @wraps(fn)
    def curried(*args, **kwargs):
        updated_fn = update_args(fn, args, kwargs)
        if is_filled(updated_fn, spec):
            return updated_fn()
        else:
            return wraps(fn)(make_curry(updated_fn, spec))
    curried.__curried__ = True
    return curried


def curry(fn: Callable):
    spec = getfullargspec(fn)
    # 只有一个参数或者没有参数时：返回方法本身
    return make_curry(fn, spec) if _need_curried_f(spec) else fn


def is_curried(f):
    return hasattr(f, '__curried__') and f.__curried__


def flip(f):
    """
    反转一个【双参数】方法的参数位置

    ps: 如果需要修改多参数方法的位置，请使用 `_` 占位符
    :param f:
    :return:
    """
    @curry
    @wraps(f)
    def flipped(a, b, *args, **kwargs):
        return f(b, a, *args, **kwargs)
    flipped.__doc__ = f'**fliped two params on head**\n{flipped.__doc__}'
    return flipped





from typing import Iterable, MutableSequence, Sequence, Callable, Type, Optional, Any

import more_itertools as mil

from jcramda.core import curry, between, flip, of, islice, not_none, concat, co, mapof, select

__all__ = (
    'append', 'prepend', 'pop', 'shift', 'update', 'adjust', 'slices',
    'chunked', 'windowed', 'padded', 'nth_or_last', 'iterate', 'split_before',
    'split_after', 'split_at', 'split_into', 'split_when', 'distribute', 'adjacent', 'dotproduct',
    'locate', 'lstrip_f', 'rstrip_f', 'strip_f', 'take', 'tabulate', 'tail', 'consume', 'nth',
    'all_eq', 'quantify', 'ncycles', 'find_true', 'iter_except', 'unique_set', 'grouper',
    'partition_f', 'update_range', 'drop', 'startswith', 'endswith', 'symdiff', 'zip_eq', 'diff'
)


@curry
def append(x, seqs: Iterable):
    return of(seqs, x)


def prepend(x, seqs: Iterable):
    return of(x, seqs)


def pop(seqs: MutableSequence, index=0):
    return seqs.pop(index)


def shift(seqs: MutableSequence):
    return seqs.pop(0)


@curry
def update(index: int, v, seqs: MutableSequence):
    if between(0, len(seqs), index):
        seqs[index] = v
    return seqs


@curry
def update_range(upset, seqs: Sequence, start=0, stop=None, step=1):
    r = list(seqs)
    for i, v in islice((start, stop, step), enumerate(r)):
        r[i] = upset(v)
    return r


@curry
def adjust(index: int, f: Callable, seqs: MutableSequence):
    if between(0, len(seqs), index):
        seqs[index] = f(seqs[index])
    return seqs


@curry
def slices(_s: tuple, seqs: Sequence):
    return seqs[slice(*_s)]


# more itertools curried
chunked = flip(mil.chunked)
windowed = flip(mil.windowed)


@curry
def padded(v, n, iterable, next_multiple=False):
    return mil.padded(iterable, v, n, next_multiple)


take = curry(mil.take)


@curry
def drop(n: int, seqs: Sequence):
    return type(seqs)(seqs[n:])


tail = curry(mil.tail)
tabulate = curry(mil.tabulate)
consume = flip(mil.consume)


@curry
def nth(idx, iterable):
    return mil.nth(iterable, int(idx))


nth_or_last = flip(mil.nth_or_last)
# 递归: (f, start) -> ``start``, ``f(start)``, ``f(f(start))``, ...
iterate = curry(mil.iterate)
split_at = flip(mil.split_at)
split_before = flip(mil.split_before)
split_after = flip(mil.split_after)
split_when = flip(mil.split_when)
split_into = flip(mil.split_into)
distribute = curry(mil.distribute)
adjacent = curry(mil.adjacent)
dotproduct = curry(mil.dotproduct)
# Yield the index of each item in *iterable* for which *pred* returns ``True``
locate = flip(mil.locate)
lstrip_f = flip(mil.lstrip)
rstrip_f = flip(mil.rstrip)
strip_f = flip(mil.strip)
all_eq = mil.all_equal
quantify = flip(mil.quantify)
ncycles = flip(mil.ncycles)
grouper = flip(mil.grouper)
partition_f = curry(mil.partition)
unique_set = curry(mil.unique_everseen)
iter_except = curry(mil.iter_except)
zip_eq = curry(mil.zip_equal)
find_true = curry(lambda f, xs: mil.first_true(xs, None, f))


@curry
def diff(s1: Sequence, s2: Sequence):
    return tuple(x for x in s1 if x not in s2)


@curry
def startswith(prefix, s: Sequence, start=None, end=None):
    return prefix == s[slice(start, end)][0:len(prefix)]


@curry
def endswith(suffix, s: Sequence, start=None, end=None):
    return suffix == s[slice(start, end)][-len(suffix):]


@curry
def symdiff(func, seq1, seq2):
    """ fetch item that func(x) just in seq1 or just in seq2
    :param func: a function
    :param seq1: Sequence
    :param seq2: Sequence
    :return tuple
    """
    rs1 = mapof(func, seq1)
    rs2 = mapof(func, seq2)
    selector = concat(
        mapof(lambda x: x not in rs2, rs1),
        mapof(lambda y: y not in rs1, rs2),
    )
    return co(
        of,
        select(selector),
        concat(seq1)
    )(seq2)



from typing import TypeVar, Callable, Union
import datetime as dt
from enum import IntEnum, auto
from ..core import curry, when, is_a


__all__ = (
    'daterange',
    'format_dt',
    'fromtimestamp',
    'FreqName',
    'locnow',
    'localize',
    'now',
    'fmt_std_date',
    'fmt_std_datetime',
    'timestamp',
    'timestamp_ms',
    'to_datetime',
    'today',
    'utcnow',
)


AnyDateTime = TypeVar('AnyDateTime', dt.date, dt.datetime)

now: Callable[[], dt.datetime] = dt.datetime.now
utcnow: Callable[[], dt.datetime] = dt.datetime.utcnow
today = dt.date.today


def locnow():
    from dateutil.tz import tzlocal
    return now(tzlocal())


def timestamp(d: dt.datetime = None):
    return int(d.timestamp() if d else now().timestamp())


def timestamp_ms(d: dt.datetime = None):
    return int(d.timestamp() if d else now().timestamp()) * 1000


@curry
def localize(_tz: dt.tzinfo, t1: dt.datetime):
    return t1.astimezone(_tz)


def fromtimestamp(tmp):
    try:
        return dt.datetime.fromtimestamp(float(tmp))
    except (TypeError, ValueError):
        return None


def to_datetime(raw):
    from dateutil.parser import parse
    if is_a((str, bytes), raw) and raw.isdigit():
        raw = int(raw)
    try:
        return when([
            (is_a((str, bytes)), parse),
            (is_a((int, float)), fromtimestamp),
        ], None)(raw)
    except (ValueError, TypeError):
        return None


@curry
def format_dt(p: str, d: AnyDateTime):
    return d.strftime(p) if d else ''


fmt_std_date = format_dt('%Y-%m-%d')
fmt_std_datetime = format_dt('%Y-%m-%d %H:%M:%S')


class FreqName(IntEnum):
    Yearly = auto()
    Monthly = auto()
    Weekly = auto()
    Daily = auto()
    Hourly = auto()
    Minutely = auto()
    Secondly = auto()


@curry
def daterange(freq: Union[int, FreqName], start_date, **kwargs):
    """
    生产一个日期列表，使用dateutil.rrule
    :param freq: FreqName
        生成的频率：年(0)、月(1)、周(2)、日(3)、小时(4)、分(5)、秒(6)
    :param start_date: 开始日期
    :param kwargs:
        @see dateutil.rrule
    :return:
    """
    from dateutil.rrule import rrule
    value = freq.value - 1 if isinstance(freq, FreqName) else freq

    return rrule(value, dtstart=start_date, **kwargs)







from more_itertools import repeatfunc, zip_equal
from functools import wraps
from jcramda.core import curry, of, map_

__all__ = (
    'applyto',
    'converge',
    'juxt',
    'until',
    'f_digest',
    'repeat_call',
    'use_with',
    'pair_call',
    'use_with',
    'always_call',
)


class applyto:
    def __init__(self, *args, **kwargs):
        self._args = args
        self._kw = kwargs

    def __call__(self, fn):
        if callable(fn):
            return fn(*self._args, **self._kw)
        return None


def juxt(*funcs):
    """
    call some functions with a same param value.
    :param funcs:
    :return: a list with functions call result
    """
    def juxt_w(*args, **kwargs):
        return map_(applyto(*args, **kwargs), funcs)

    return juxt_w


@curry
def until(pred, funcs, v, *args, **kwargs):
    for func in funcs:
        r = func(v, *args, **kwargs)
        if pred(r):
            return r
    return None


def f_digest(f):
    from inspect import signature
    from pickle import dumps
    from .text import hexdigest
    return hexdigest('sha256', dumps(signature(f)))


@curry
def converge(after_f, funcs, value):
    """
    (g, [f1, f2, ... fn], value) -> g(f1(value), f2(value), ... fn(value))
    :param after_f: 最终处理函数
    :param funcs: 函数列表
    :param value: 要处理的值
    :return:
    """
    return after_f(*juxt(*funcs)(value))


repeat_call = curry(repeatfunc)


@curry
def pair_call(funcs, v, *args):
    values = of(v, args)
    return (f(*of(x)) for (f, x) in zip_equal(funcs, values))


@curry
def use_with(after_f, funcs, v, *args):
    return after_f(*pair_call(funcs, v, *args))



def always_call(f, *args, **kwargs):
    def _just_call(__):
        return f(*args, **kwargs)
    return wraps(f)(_just_call)


@curry
def tap(f, v):
    f(v)
    return v




from enum import Enum

from jcramda.core import curry, co, attr, bind, index, is_a
from jcramda.base.sequence import find_true, nth
from jcramda.base.comparison import attr_eq

__all__ = (
    'is_enum',
    'members',
    'valueof',
    'nameof',
    'e_index_of',
    'e_index_by',
    'enum_name',
    'enum_value',
)

is_enum = is_a(Enum)


members = co(
    bind('values'),
    attr('__members__'),
)


@curry
def valueof(v, e):
    return co(
        find_true(attr_eq('value', v)),
        members,
    )(e)


@curry
def nameof(name, e):
    return co(
        find_true(attr_eq('name', name)),
        members,
    )(e)


# (member, Enum) -> int
e_index_of = curry(lambda m, e: index(m, members(e)))

e_index_by = curry(lambda i, e: nth(i)(members(e)))


def enum_name(e):
    return e.name if is_enum(e) else e


def enum_value(e):
    return e.value if is_enum(e) else e



from typing import Dict, List, Tuple, AnyStr, Callable, Iterable, Mapping, Sequence
from jcramda.core import curry, flip, co, lt, le, gt, ge, eq, attr, is_a, all_, not_a

__all__ = (
    'ilen_lt', 'ilen_le', 'ilen_eq', 'ilen_gt', 'ilen_ge',
    'len_lt', 'len_le', 'len_eq', 'len_gt', 'len_ge', 'is_zero', 'is_nan',
    'attr_eq', 'is_a_dict', 'is_a_list', 'is_a_tuple', 'is_a_str', 'is_a_func',
    'is_a_int', 'is_iter', 'is_a_mapper', 'is_seq', 'is_simple_iter', 'nostr_seq',
)


ilen_lt = curry(lambda n, x: len(x) < n)
ilen_le = curry(lambda n, x: len(x) <= n)
ilen_eq = curry(lambda n, x: len(x) == n)
ilen_gt = curry(lambda n, x: len(x) > n)
ilen_ge = curry(lambda n, x: len(x) >= n)

len_lt = curry(lambda b, a: len(a) < len(b))
len_le = curry(lambda b, a: len(a) <= len(b))
len_eq = curry(lambda b, a: len(a) == len(b))
len_gt = curry(lambda b, a: len(a) > len(b))
len_ge = curry(lambda b, a: len(a) >= len(b))

is_zero = eq(0)


def is_nan(x):
    return str(x).lower() in ('nan', 'null', 'none')


@curry
def attr_eq(attr_name, value, obj):
    return co(
        eq(value),
        attr(attr_name),
    )(obj)


# typing

is_a_dict = is_a(Dict)
is_a_mapper = is_a(Mapping)
is_a_tuple = is_a(Tuple)
is_a_list = is_a(List)
is_a_str = is_a(AnyStr)
is_a_func = is_a(Callable)
is_a_int = is_a(int)
is_iter = is_a(Iterable)
is_simple_iter = all_([is_iter, not_a((bytes, str, Mapping))])
is_seq = is_a(Sequence)
nostr_seq = all_([is_seq, not_a((bytes, str))])




import hashlib
import re
from random import choices
from typing import Iterable, AnyStr
from jcramda.core import (curry, bind, co, chain, repeat, always)
from jcramda.base.sequence import update_range, split_before
from string import ascii_lowercase


__all__ = (
    'center', 'encode', 'expandtabs', 'decode', 'find', 'join',
    'format_map', 'ljust', 'lstrip', 'partition', 'replace', 'rfind', 'rindex', 'rjust',
    'rpartition', 'rsplit', 'rstrip', 'strcount', 'sformat', 'split', 'strip', 'translate', 'zfill',
    # custom functions
    'first_lower', 'hex_token', 'url_safe_token', 'hex_uuid', 'camelcase', 'camelcase_to',
    'rand_txt', 'repeat_txt', 'mask', 'mask_except', 'hexdigest', 'b64_encode', 'b64_decode',
    'b64_urlsafe_encode', 'b64_urlsafe_decode', 'search', 'match', 'fullmatch', 'resub', 'resubn',
    'finditer', 'findall', 'url_join',
)

# string method curried ========================================


@curry
def replace(old_str, new_str, s, _count=-1):
    return s.replace(old_str, new_str, count=_count)


@curry
def sformat(tmpl: str, s, *args, **kwargs):
    return tmpl.format(s, *args, **kwargs)


@curry
def strip(s: str, chars=None):
    return s.strip(chars)


@curry
def lstrip(s: str, chars=None):
    return s.lstrip(chars)


@curry
def rstrip(s: str, chars=None):
    return s.rstrip(chars)


@curry
def strcount(rs, s: str, start=None, end=None):
    return s.count(rs, start, end)


def encode(s: AnyStr, errors='ignore', encoding='utf8'):
    if isinstance(s, str):
        return s.encode(encoding, errors)
    return s


def decode(bs: AnyStr):
    if isinstance(bs, bytes):
        return bs.decode()
    return bs


@curry
def expandtabs(tab_size, s: str):
    return s.expandtabs(tab_size)


@curry
def find(rs, s: str, start=None, end=None):
    return s.find(rs, start, end)


@curry
def rfind(rs, s: str, start=None, end=None):
    return s.rfind(rs, start, end)


@curry
def join(sep: str, seqs: Iterable):
    return sep.join([str(x) for x in seqs if x is not None])


@curry
def ljust(width, s: str, fillchar=None):
    return s.ljust(width, fillchar)


@curry
def rjust(width, s: str, fillchar=None):
    return s.rjust(width, fillchar)


@curry
def translate(opts: dict, s: str):
    return s.translate(str.maketrans(opts))


@curry
def rindex(subs, s: str, start=None, end=None):
    return s.rindex(subs, start, end)


@curry
def split(sep, s: str, limit=-1):
    return s.split(sep, limit)


@curry
def rsplit(sep, s: str, limit=-1):
    return s.rsplit(sep, limit)


def center(width: int, s: str, fill_char=' '):
    return s.center(width, fill_char)


format_map = curry(lambda mapping, s: s.format_map(mapping))
partition = curry(lambda sep, s: s.partition(sep))
rpartition = curry(lambda sep, s: s.rpartition(sep))
zfill = curry(lambda length, s: s.zfill(length))


# custom =======================================================
def first_lower(s: str):
    return s[0:1].lower() + s[1:]


def hex_token(size):
    from secrets import token_hex
    return token_hex(size)


def url_safe_token(size):
    from secrets import token_urlsafe
    return token_urlsafe(size)


def hex_uuid():
    from uuid import uuid4
    return uuid4().hex


def camelcase(s: AnyStr):
    """
    下划线转换驼峰
    :param s:
    :return:
    """
    return co(
        first_lower,
        join(''),
        update_range(str.capitalize, start=1),
        split('_')
    )(s)


@curry
def camelcase_to(sep, s: AnyStr, trans_f=str.lower):
    return co(
        trans_f,
        join(sep),
        chain(join('')),
        split_before(str.isupper)
    )(s)


def rand_txt(length, char_set=ascii_lowercase):
    return join('')(choices(char_set, k=length))


@curry
def repeat_txt(n, s):
    return join('', repeat(n, s))


@curry
def mask(start, stop, raw: str, char='*'):
    return co(
        join(''),
        update_range(always(char), start=start, stop=stop)
    )(raw)


@curry
def mask_except(_head: int, _tail: int, s: str, char='*'):
    return s[0:_head] + repeat_txt(len(s) - _head - _tail, char) + s[-_tail:]


@curry
def hexdigest(algorithm, raw: AnyStr, length=32):
    """
    支持以下算法
    ('md5', 'sha1', 'sha224', 'sha256', 'sha384', 'sha512', 'blake2b',
     'blake2s', 'sha3_224', 'sha3_256', 'sha3_384', 'sha3_512', 'shake_128', 'shake_256')
    :param algorithm:
    :param raw:
    :param length:
    :return:
    """
    spec = bind(algorithm, encode(raw))
    try:
        return spec(hashlib).hexdigest()
    except AttributeError:
        raise RuntimeError(f'not supported algorithm: {algorithm}')
    except TypeError:
        return spec(hashlib).hexdigest(length)


def b64_encode(s: AnyStr):
    import base64
    return co(
        decode,
        base64.b64encode,
        encode,
    )(s)


def b64_urlsafe_encode(s: AnyStr):
    import base64
    return co(
        decode,
        base64.urlsafe_b64encode,
        encode,
    )(s)


def b64_decode(s: AnyStr):
    import base64
    return co(
        decode,
        base64.b64decode,
    )(s)


def b64_urlsafe_decode(s: AnyStr):
    import base64
    return co(
        decode,
        base64.urlsafe_b64decode,
    )(s)


search = curry(re.search)
match = curry(re.match)
fullmatch = curry(re.fullmatch)
resub = curry(re.sub)
resubn = curry(re.sub)
finditer = curry(re.finditer)
findall = curry(re.findall)

def url_join(*args):
    first_char = '/' if args[0][0] == '/' else ''
    return first_char + join('/', map(lambda s: s.strip('/'), args))



from functools import reduce, partial
from typing import Callable

__all__ = (
    'compose',
    'pipe',
    'co',
    'partial',
    'break_if',
)


class ComposeBorker:
    def __init__(self, last_result, else_=None):
        self._last_result = last_result
        self._else_f = else_

    @property
    def last_result(self):
        return self._else_f(self._last_result) if callable(self._else_f) else self._last_result



def compose(*fns: Callable):
    assert len(fns) > 1, 'compose must have less two functions.'
    def composed_func(*args, **kwargs):
        f, *funcs = reversed(fns)
        result = f(*args, **kwargs)        
        for func in funcs:       
            if isinstance(result, ComposeBorker):
               return result.last_result
            result = func(result)            
        return result

    return composed_func


co = compose


def pipe(*funcs: Callable):
    return compose(*reversed(funcs))


def break_if(pred, else_=None):
    def borker(result):
        if pred(result):
            return ComposeBorker(result, else_)
        return result
    return borker



import itertools as its
from functools import reduce as _reduce
from typing import Iterable, Callable, Mapping, Tuple

from more_itertools import (
    with_iter,
    intersperse as _intersperse,
    consume,
    side_effect,
    ilen,
    always_reversible as reverse,
    replace as _replace,
    one as _one,
    filter_except as _fet,
    map_except as _met,
)

from ._curry import curry, flip
from .compose import co

__all__ = (
    'aof',
    'of',
    'flatten',
    'one',
    'cycle',
    'count',
    'some',
    'select',
    'fold',
    'chain',
    'first',
    'last',
    'maps',
    'map_',
    'map_except',
    'reduce_',
    'mapof',
    'each',
    'foreach',
    'fmap',
    'fmapof',
    'filter_',
    'filter_of',
    'filter_group',
    'filter_not',
    'filter_except',
    'dropwhile',
    'takewhile',
    'product',
    'permute',
    'combine',
    'zip_',
    'islice',
    'with_iter',
    'ilen',
    'repeat',
    'reverse',
    'ireplace',
    'map_reduce',
    'scan',
    'groupby',
)


@curry
def aof(*args):
    return tuple(*args,)


# noinspection PyArgumentList
@curry
def of(*args, cls=tuple):
    """
    将传入的参数平铺成一个tuple
    :param cls: 类，可以是 （tuple, list, set, etc...）或其他衍生可迭代容器类或方法，默认: tuple
    :param args: Iterable
    :return: cls指定的迭代类型
    """
    data = []
    for x in args:
        if isinstance(x, Iterable) and not isinstance(x, Mapping):
            data += list(x)
        else:
            data.append(x)
    return cls(data)


@curry
def flatten(*args):
    if len(args) == 1 and not isinstance(args[0], Iterable):
        return args[0]
    from more_itertools import collapse
    return *collapse(args),


def one(iterable):
    """ 如果传入的迭代器仅有一个元素，则返回这个元素，否则返回迭代器的结果（Tuple） """
    r = tuple(iterable) if isinstance(iterable, Iterable) else iterable
    try:
        return _one(r)
    except (TypeError, IndexError, ValueError):
        return r


def first(iterable):
    return next(iter(iterable), None)


def last(iterable):
    try:
        return iterable[-1]
    except (TypeError, AttributeError, KeyError):
        from collections import deque
        return deque(iterable, maxlen=1)[0]


@curry
def fold(func, init, it):
    """
    统合数据，即传统的 reduce 方法
    :param func:
    :param init:
    :param it:
    :return:
    """
    return _reduce(func, it, init)


reduce_ = fold


@curry
def map_(func, it):
    return (func(x) for x in it)


@curry
def mapof(func, it):
    return of(*(func(x) for x in it))


# maps: (fun, (iter1, iter2...iterN)) ->
@curry
def maps(f: Callable, it, *args):
    return its.starmap(f, tuple((it, *args)))


@curry
def map_except(func, exceptions, iterable):
    return _met(func, iterable, *exceptions)


@curry
def islice(rng: Tuple[int], iterate):
    return its.islice(iterate, *rng)


# a side effect each: (f, seqs, chunk_size, before, after) -> seqs
each = curry(side_effect)


@curry
def foreach(func, seqs, chunk_size=None, before=None, after=None):
    consume(side_effect(func, seqs, chunk_size, before, after))
    return seqs


@curry
def fmap(func, seqs):
    return its.chain(*map(func, seqs))


@curry
def fmapof(func, seqs):
    return tuple(its.chain(*map(func, seqs)))


@curry
def repeat(n, x):
    return its.repeat(x, n)


dropwhile = curry(its.dropwhile)
takewhile = curry(its.takewhile)


@curry
def filter_(func, seqs):
    return (x for x in seqs if func(x))


@curry
def filter_of(f, seqs):
    return of(x for x in seqs if f(x))


filter_not = curry(its.filterfalse)


@curry
def filter_group(pred: Callable[..., bool], seqs):
    truth = []
    false = []
    for x in seqs:
        if pred(x):
            truth.append(x)
        else:
            false.append(x)
    return tuple(truth), tuple(false)


@curry
def filter_except(pred, exceptions, iterable):
    """Yield the items from *iterable* for which the *validator* function does
    not raise one of the specified *exceptions*.

    *validator* is called for each item in *iterable*.
    It should be a function that accepts one argument and raises an exception
    if that item is not valid.

    >>> iterable = ['1', '2', 'three', '4', None]
    >>> of(filter_except(int, (ValueError, TypeError))(iterable))
    ('1', '2', '4')

    If an exception other than one given by *exceptions* is raised by
    *validator*, it is raised like normal.
    """
    return _fet(pred, iterable, *exceptions)


@curry
def some(pred, iterable) -> bool:
    return isinstance(iterable, Iterable) and len([x for x in iterable if pred(x)]) > 0


@curry
def product(iter1, iter2, *args, r=1):
    return its.product(iter1, iter2, *args, repeat=r)


@curry
def product_map(func, iter1, iter2, *args):
    return (func(x) for x in its.product(iter1, iter2, *args))


@curry
def permute(seqs, r):
    """
    排列
    :param seqs: Iterable[_T]
    :param r: int
        如果 r 是0则返回全排列
    :return: Iterable[Tuple[_T]]
    """
    return its.permutations(seqs, r) if r else its.permutations(seqs)


@curry
def combine(seqs, r):
    """
    组合
    :param seqs:
    :param r:
    :return:
    """
    return its.combinations(seqs, r)


@curry
def zip_(fill_value, seq, *seqs):
    return its.zip_longest(seq, *seqs, fillvalue=fill_value)


@curry
def groupby(func, iterable):
    return its.groupby(iterable, func),


@curry
def chain(*args):
    funcs = tuple(reverse(f for f in args if callable(f)))
    first_func = first(funcs)
    if first_func is None:
        return flatten(*args)
    return co(one, flatten,
              map_(lambda x: fold(lambda r, f: f(r, x), first_func(x), funcs[1:])),
              aof,
              )


# 等距插入固定元素 (e, iterable, n=1) -> Iterable
#         >>> list(intersperse('!', [1, 2, 3, 4, 5]))
#         [1, '!', 2, '!', 3, '!', 4, '!', 5]
intersperse = curry(_intersperse)


@curry
def ireplace(pred, sub, iterable, _count=None, window_size=1):
    return _replace(iterable, pred, sub, _count, window_size)


@curry
def map_reduce(key: Callable, iterable: Iterable, emit=None, reducer=None):
    from more_itertools import map_reduce
    return map_reduce(iterable, key, emit, reducer)


@curry
def scan(func, init, iterable):
    result = [init]
    r = init
    for x in iterable:
        r = func(r, x)
        result.append(r)
    return iter(result)


select = flip(its.compress)
count = curry(its.count)
cycle = its.cycle




from setuptools import setup, find_packages


# readme = open('Readme.md', 'r').read()


setup(
    name='jcramda',
    version='1.0.3',
    author='Jochen He',
    author_email='thjl@hotmail.com',
    description='functional programming methods',
    long_description='',
    long_description_content_type='text/markdown',
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'more-itertools',
        'pytz',
        'python-dateutil',
    ],
    python_requires='>=3.6'
)

